"""
Thomas Wick, Sebastian Kinnewig
Leibniz University Hannover
Institute of Applied Mathematics
Germany

https://www.pec3.org/events/school2022/

Exercise Class 1 - Iterative Solvers and Gradient Decent Methods

Exercise 1 - Basic Iterative Solvers based on Fixed-Point Schemes
"""

"""
===== PACKAGES =====
First we begin by loading some usefull packages: numpy and matplotlib
"""
# Basic math tools 
import numpy as np
# (Installation for Windows: 
#    open the cmd and type into it: "python -m pip install numpy")
# (Installation for Linux: 
#    first install pip3, therefore open a Terminal and type "apt install pip3"
#    then type into the terminal "pip3 install numpy")


# Library to create plots 
import matplotlib.pyplot as plt 
# (Installation for Windows: 
#    open the cmd and type into it: "python -m pip install matplotlib")
# (Installation for Linux: 
#    open a terminal and type into it: "pip3 install matplotlib")


""" 
===== Fixed section ===
In this section, we create the system matrix A from the Poisson problem, 
which was introduced on the exercise sheet and the right-hand side for 
the Poisson problem f, also as introduced in equation (1) on the exercise 
sheet
"""
# This function is used to create the system matrix A
def laplace_matrix( n, L = 1.0 ):
  # initialize an (n + 2) x (n + 2) matrix with zeros
  A = np.zeros((n + 2, n + 2))

  # compute h
  h = L / ( n + 1)

  # fill in the matrix (remark: we skip the first and the last line of the matrix)
  for i in range( 1, n + 1 ):
    # diagonal
    A[i][i] = 2 / h ** 2
    # fill the off diagonals
    A[i][i - 1] = - 1 / h ** 2
    A[i][i + 1] = - 1 / h ** 2

  # fill in the boundary values
  A[0][0] = 1 / h ** 2
  A[n + 1][n + 1] = 1 / h ** 2

  return A

# This function is used to create the right-hand side
def laplace_rhs( n ):
  # initialze a vector of length n+2 with zero
  f = np.zeros(n + 2)
  
  # fill the vector (remark: we skip the first and the last line of the vector)
  for i in range( 1, n + 1 ):
    f[i] = -1.0

  return f
  
# Create a plot
def create_plot( x ):
  plt.plot(x)
  plt.ylabel('u(x)')
  plt.xlabel('x')
  plt.show()
  


""" 
===== Bonus: Warm Up =====
As a small warm-up exercise, we want to implement here the decomposition  
of the matrix A into the lower triangular matrix L, the diagonal matrix D 
and the upper triangular matrix U as explained in the lecture on slide 31
"""

def lower_triangular( A ):
  # get the size of A
  size = len( A )

  # create a new output matrix
  L = np.zeros((size, size))

  # copy the corresponding entries
  # TODO

  return L;

def upper_triangular( A ):
  # get the size of A
  size = len( A )

  # create a new output matrix
  U = np.zeros((size, size))

  # copy the corresponding entries
  # TODO

  return U;

def diagonal( A ):
  # get the size of A
  size = len( A )

  # create a new output matrix
  D = np.zeros((size, size))

  # copy the corresponding entries
  # TODO

  return D

# Let us create a small test case, where we can test if our functions
# defined above are working correctly

# Remark: When you begin with this exercise please uncomment the following lines
#         once you are done, the best practice is to command out the lines again

#test = np.array([[1,2,3],[4,5,6],[7,8,9]])
#print( "Decompose the matrix: ")
#print( test )
#print( "into the lower triangular matrix L =" )
#print( lower_triangular( test ) )
#print( "the upper triangular matrix U =" )
#print( upper_triangular( test ) )
#print( "and the diaginal matrix D =")
#print( diagonal( test ) )



"""
===== Exercise 1 (a): Jacobi Method =====
Replace the TODO with the iteration step of the Jacobi method (in index-notation)
as provided on slide 32 or the exercise sheet.
"""
def jacobi_sum(A, x_old, i):
  n = len( x_old )

  result = 0
  for j in range(n):
    if i != j:
      result += A[i][j] * x_old[j]

  return result

def jacobi(A, f, epsilon = 1e-5, max_iterations = 1000):
  # get the size of A
  size = len( A )

  # iteration start:
  # create an initial guess x0
  x_old = np.zeros( size )
  x_new = np.zeros( size )
    
  # iteration step
  for k in range(max_iterations):

    # update
    x_old = x_new.copy()

    for i in range(size):
      # compute the new x vector (with index notation)
      #TODO

    # check for convergence
      if np.sqrt( np.dot(x_old - x_new, x_old - x_new) ) < epsilon:
        print("Converged after ", k, " steps")
        break


  # return the result
  return x_new

# We want to solve the following LES:
# A.x = f

# Remark: When you begin with this exercise please uncomment the following lines
#n = 4                  # number of discretization points
#A = laplace_matrix(n)  # (Given) The system matrix from LES
#f = laplace_rhs(n)     # (Given) The right hand side from the LES
#x = jacobi(A, f)       # Compute x with jacobi method you completed above
#create_plot(x)

"""
===== Exercise 1 (b) Gauss-Seidel Method =====
Implement the Gauss-Seidel method using the index-notation as provided on slide 33
or the exercise sheet. You can use the following layout as a starting point.
"""
def gauss_sum_part_1(A, x_new, i):
  result = 0

  #TODO

  return result

def gauss_sum_part_2(A, x_old, i):
  result = 0

  #TODO

  return result

def gauss(A, f, epsilon = 1e-5, max_iterations = 1000):
  x_new = np.zeros( size )
    
  #TODO

  return x_new

# We want to solve the following LES:
# A.x = f

# Remark: When you begin with this exercise please uncomment the following lines
#n = 4                  # number of discretization points
#A = laplace_matrix(n)  # (Given) The system matrix from LES
#f = laplace_rhs(n)     # (Given) The right hand side from the LES
#x = gauss(A, f)        # Compute x with the gauss-seidel method you implemented above
#create_plot(x)

"""
===== Exercise 1 (c) Comparison Between the Jacobi and the Gauss-Seidel Method =====
Consider different n (for example n = 4, 8, 16) and compare the results between the two different methods. Especially compare the number of iteration steps of both methods.
"""
